/*****************************************************************************

This source code is made available to you under the terms of the Open
Development License.  This Package should be accompanied by a file that
contains the associated licensing terms.  If you are unable to find this file,
please refer to its location on the Internet at www.rocklyte.com/license.html.

******************************************************************************

###PROGRAM###
Name:      Mini Player
Author:    Rocklyte systems
Copyright: Rocklyte Systems, 2003-2004.  All rights reserved.
###END###

TODO:

+ Volume control
+ Soundtrack positioning slider.
+ Left/Right speaker panning.

*****************************************************************************/

//#define DEBUG

#include <pandora/system/all.h>
#include <pandora/graphics/all.h>
#include <pandora/tools/filedialog.h>
#include <pandora/modules/strings.h>
#include <pandora/modules/filesystem.h>
#include <pandora/modules/render.h>
#include <pandora/audio/all.h>
#include <pandora/data/xml.h>
#include <pandora/startup.h>

STRING ProgName      = "Mini Player";
STRING ProgAuthor    = "Rocklyte Systems";
STRING ProgDate      = "January 2004";
STRING ProgCopyright = "Copyright Rocklyte Systems © 2003-2004.  All rights reserved.";
LONG   ProgDebug = 0;
FLOAT  ProgKernelVersion = 3.3;

#define PROGVERSION ((FLOAT)1.1)

enum {
   STATE_STOPPED,
   STATE_PAUSED,
   STATE_PLAYING
};

extern struct KernelBase *KernelBase;
static struct StringsBase *StringsBase;
static struct FileSystemBase *FileSystemBase;
static struct RenderBase *RenderBase;
static struct Font *glFont = NULL;
static OBJECTPTR modStrings = NULL, modFileSystem = NULL, modRender = NULL;
static OBJECTPTR glToolbar = NULL, glCurrent = NULL;
static OBJECTPTR glFileDialog = NULL;
static OBJECTPTR glPlaylistView = NULL;
static OBJECTID glWindowID = NULL, glPlaylistViewID = NULL;
static OBJECTPTR glTask;
static OBJECTID glOutputID;
static OBJECTID glPLAdd = NULL, glPLDelete = NULL;
static LONG glState = STATE_STOPPED;
static LONG glStringWidth = 0, glPanelWidth = 0;
static UBYTE glString[160] = "";
static UBYTE glFile[256] = "";
static UBYTE glTitle[256] = "";
static struct { STRING File; STRING Title; LONG Length; } glPlaylist[256];
static WORD glPlayIndex = 0, glTrack = 0, glScrollX = 0;
static UBYTE glScrollDirection = 0;
static BYTE glAddFiles = FALSE;

#define PANEL_MARGIN 6

/*****************************************************************************
** Prototypes
*/

static ERROR PROGRAM_ActionNotify(OBJECTPTR, struct acActionNotify *);
static ERROR PROGRAM_Timer(OBJECTPTR, struct acTimer *);

static void AboutWindow(void);
static void AddPlayFile(STRING File);
static void AddViewXML(WORD Index);
static void ClearPlaylist(void);
static void DrawOutput(APTR, struct Render *, struct Bitmap *);
static STRING GetTitle(OBJECTPTR, STRING);
static STRING GetXMLAttribute(struct XMLTag *, STRING);
static LONG GetXMLAttributeInt(struct XMLTag *, STRING);
static void LoadTrack(WORD);
static ERROR MessageHandler(struct Message *);
static ERROR OpenFile(STRING);
static void Pause(void);
static void PlayAudio(void);
static void PlaylistWindow(void);
static void PlayPrevious(void);
static void PlayTrack(WORD);
static void RemovePlayFile(LONG Index);
static void SeekAudio(DOUBLE);
static void SetString(STRING);
static void StopPlaying(void);

/*****************************************************************************
** Toolbar definition.
*/

enum {
   IDX_PREVIOUS=0,
   IDX_PLAY,
   IDX_PAUSE,
   IDX_STOP,
   IDX_NEXT,
   IDX_BREAK_1,
   IDX_OPEN,
   IDX_PLAYLIST
};

static UBYTE strToolbar[] = "\
<item text=\"Previous\" tip=\"Previous File\" icon=\"items/player_prev\"/>\
<item text=\"Play\" tip=\"Play Audio\" icon=\"items/player_play\"/>\
<item text=\"Pause\" tip=\"Pause Audio\" icon=\"items/player_pause\"/>\
<item text=\"Stop\" tip=\"Stop Audio\" icon=\"items/player_stop\"/>\
<item text=\"Next\" tip=\"Next File\" icon=\"items/player_next\"/>\
<break/>\
<item text=\"Open File\" tip=\"Open File(s)\" icon=\"items/player_eject\"/>\
<item text=\"Playlist\" tip=\"View Playlist\" icon=\"office/documents\"/>";

static UBYTE glHelp[] = "\
MiniPlayer Options\
\
-help   This help page.\
-file   Reference to an audio file that you want to play at startup.\
";

#ifdef DEBUG
static void PrintPlaylist(void)
{
   WORD i;
   for (i=0; i < glPlayIndex; i++) {
      DPrintF("Playlist:","%.2d: %s, %s", i, glPlaylist[i].Title, glPlaylist[i].File);
   }
}
#else
#define PrintPlaylist()
#endif

/*****************************************************************************
** Internal: Failure()
*/

ERROR Failure(STRING Message)
{
   WORD i;

   if (modStrings)         acFree(modStrings);
   if (modFileSystem)      acFree(modFileSystem);
   if (modRender)          acFree(modRender);
   if (glFont)             acFree(glFont);

   for (i=0; i < glPlayIndex; i++) {
      if (glPlaylist[i].File) FreeMemory(glPlaylist[i].File);
      if (glPlaylist[i].Title) FreeMemory(glPlaylist[i].Title);
      glPlaylist[i].File = NULL;
   }

   if (Message) {
      print(Message);
      return(ERR_Failed);
   }
   else return(ERR_Okay);
}

/*****************************************************************************
** Main.
*/

ERROR program(void)
{
   struct RenderInfo info;
   struct Render *render;
   LONG topmargin, bottommargin, rightmargin, leftmargin, i, tool_height, width, height;
   OBJECTPTR window, object;
   OBJECTID parent_id;
   STRING *args;
   APTR *actions;

   SetResource(RES_MESSAGEHANDLER, (LARGE)&MessageHandler);

   glTask = CurrentTask();

   /*** Load required modules ***/

   if (CreateObject(ID_MODULE, NULL, &modStrings, NULL,
      FID_Name|TSTRING,   "strings",
      FID_Version|TFLOAT, MODVERSION_STRINGS,
      TAGEND) IS ERR_Okay) {
      if (GetField(modStrings, FID_ModBase, FT_POINTER, &StringsBase) != ERR_Okay) {
         return(ERR_GetField);
      }
   }
   else return(ERR_CreateObject);

   if (CreateObject(ID_MODULE, NULL, &modFileSystem, NULL,
      FID_Name|TSTRING,   "filesystem",
      FID_Version|TFLOAT, MODVERSION_FILESYSTEM,
      TAGEND) IS ERR_Okay) {
      if (GetField(modFileSystem, FID_ModBase, FT_POINTER, &FileSystemBase) != ERR_Okay) {
         return(ERR_GetField);
      }
   }
   else return(ERR_CreateObject);

   if (CreateObject(ID_MODULE, NULL, &modRender, NULL,
      FID_Name|TSTRING,   "render",
      FID_Version|TFLOAT, MODVERSION_RENDER,
      TAGEND) IS ERR_Okay) {
      if (GetField(modRender, FID_ModBase, FT_POINTER, &RenderBase) != ERR_Okay) {
         return(ERR_GetField);
      }
   }
   else return(ERR_CreateObject);

   /*** Intercept some actions that are required for our program ***/

   GetField(glTask, FID_Actions, FT_POINTER, &actions);
   actions[AC_ActionNotify] = PROGRAM_ActionNotify;
   actions[AC_Timer] = PROGRAM_Timer;

   SubscribeTimer(glTask->UniqueID, 100); /* 1/10 second updates */

   /*** Process any arguments passed to our program ***/

   if ((GetField(glTask, FID_ArgsList, FT_POINTER, &args) IS ERR_Okay) AND (args)) {
      for (i=0; args[i]; i++) {
         if (((StrCompare(args[i], "-file", NULL, STR_MATCHLENGTH) IS ERR_Okay) OR
              (StrCompare(args[i], "-location", NULL, STR_MATCHLENGTH) IS ERR_Okay))
             AND (args[i+1])) {
            StrCopy(args[i+1], glFile, sizeof(glFile));
         }
         else if ((StrCompare(args[i], "-help", NULL, STR_MATCHLENGTH) IS ERR_Okay) OR
                  (StrCompare(args[i], "--help", NULL, STR_MATCHLENGTH) IS ERR_Okay)) {
            print(glHelp);
         }
      }
   }

   if (CreateObject(ID_WINDOW, NULL, &window, &glWindowID,
      FID_Title|TSTRING,      ProgName,
      FID_MinWidth|TLONG,     178,
      FID_MinHeight|TLONG,    54,
      FID_MaxWidth|TLONG,     178,
      FID_MaxHeight|TLONG,    54,
      FID_Icon|TSTRING,       "programs/musicplayer",
      TAGEND) IS ERR_Okay) {

      /*** Move the window to the lower-right position of the display ***/

      GetFields(window, FID_Width|TLONG,  &width,
                        FID_Height|TLONG, &height,
                        FID_Parent|TLONG, &parent_id,
                        TAGEND);

      if (GetRenderInfo(parent_id, &info, sizeof(info)) IS ERR_Okay) {
         if (AccessObject(parent_id, 4000, (OBJECTPTR *)&render) IS ERR_Okay) {
            ActionTags(AC_MoveToPoint, window,
               (DOUBLE)(info.Width - render->RightMargin - width - 20),
               (DOUBLE)(info.Height - render->BottomMargin - height - 20),
               0.0,
               MTF_X|MTF_Y);
            ReleaseObject(render);
         }
      }

      SubscribeActionTags(window, glTask,
         AC_Free,
         TAGEND);

      GetFields(window, FID_LeftMargin|TLONG,   &leftmargin,
                        FID_TopMargin|TLONG,    &topmargin,
                        FID_BottomMargin|TLONG, &bottommargin,
                        FID_RightMargin|TLONG,  &rightmargin,
                        TAGEND);
      ReleaseObject(window);
   }
   else return(Failure("Could not open a window for this application."));

   if (CreateObject(ID_FONT, NULL, &glFont, NULL,
      FID_Face|TTRANS, "[gl_fonts.face]",
      TAGEND) != ERR_Okay) return(Failure("Failed to create font."));

   /*** Create the toolbar ***/

   if (CreateObject(ID_TOOLBAR, NULL, &glToolbar, NULL,
      FID_Owner|TLONG,  glWindowID,
      FID_XCoord|TLONG,  leftmargin,
      FID_YCoord|TLONG,  topmargin,
      FID_XOffset|TLONG, rightmargin,
      TAGEND) IS ERR_Okay) {

      acDataChannel(glToolbar, NULL, DATA_XML, 1, strToolbar, StrLength(strToolbar), 1);

      GetField(glToolbar, FID_Height, FT_LONG, &tool_height);

      SubscribeAction(glToolbar, AC_Activate, glTask);

      acShow(glToolbar);
   }
   else return(Failure("Failed to create the application toolbar."));

   /*** Create a render area for drawing graphics inside the window ***/

   if (CreateObject(ID_RENDER, NULL, &object, &glOutputID,
      FID_Name|TSTRING,    ProgName,
      FID_Container|TLONG, glWindowID,
      FID_XCoord|TLONG,    leftmargin,
      FID_YCoord|TLONG,    topmargin + tool_height + 6,
      FID_XOffset|TLONG,   rightmargin,
      FID_YOffset|TLONG,   bottommargin,
      FID_Region|TLONG,    TRUE,
      TAGEND) IS ERR_Okay) {

      ActionTags(MT_SetRenderCallback, object, glTask->UniqueID, DrawOutput);

      GetField(object, FID_Width, FT_LONG, &glPanelWidth);

      acShow(object);
      ReleaseObject(object);
   }
   else return(Failure("Failed to create rendering area."));

   /*** Generate the file dialog control ***/

   if (CreateObject(ID_FILEDIALOG, NULL, &glFileDialog, NULL,
         FID_FilterList|TSTRING, "All Files;*;MP3 Music Files;*.mp3;Digital Music Files (mod, xm, s3m...);*.mod|*.s3m|*.xm|*.it;Sound Files;*.wav",
         FID_Title|TSTRING,      "Select Audio File(s)",
         FID_Static|TLONG,       TRUE,
         FID_Location|TSTRING,   glFile,
         FID_Flags|TLONG,        FDF_MULTISELECT,
         TAGEND) IS ERR_Okay) {

      SubscribeAction(glFileDialog, AC_Activate, glTask);
   }

   {
      UBYTE str[200];
      StrFormat(str, sizeof(str), "Mini Player %.1f © Rocklyte Systems 2003-2004", PROGVERSION);
      SetString(str);
   }

   /*** Show the Mini Player window ***/

   acShowID(glWindowID);

   if (glFile[0]) {
      if (OpenFile(glFile) IS ERR_Okay) {
         AddPlayFile(glFile);
         PlayAudio();
      }
   }

   /*** MAIN LOOP:  Refer to our MessageHandler() routine to see what goes on in here ***/

   ProcessMessages(NULL, MF_WAIT);

   /*** DIIIEEEEEE!!! ***/

   Failure(NULL);
   return(ERR_Okay);
}

/*****************************************************************************
** Function: MessageHandler()
**
** This function is called by the system whenever it decides that we have
** messages that need to be processed.
*/

static ERROR MessageHandler(struct Message *Message)
{
   if (Message->Type IS MSGID_QUIT) {
      /* Returning a terminate code will force all further ProcessMessages()
      ** calls to return.
      */
      return(ERR_Terminate);
   }
   else return(ERR_Failed);
}

/*****************************************************************************
** Program: ActionNotify()
*/

ERROR PROGRAM_ActionNotify(OBJECTPTR Self, struct acActionNotify *Args)
{
   if (Args->Error != ERR_Okay) return(ERR_Okay);

   if ((Args->ActionID IS AC_Free) AND (Args->ObjectID IS glWindowID)) {
      SendMessage(NULL, MSGID_QUIT, NULL, NULL, NULL);
   }
   else if ((Args->ActionID IS AC_Activate) AND (Args->ObjectID IS glToolbar->UniqueID)) {
      LONG index;

      GetField(glToolbar, FID_Index, FT_LONG, &index);

      switch(index) {
         case IDX_PREVIOUS:
            PlayPrevious();
            break;

         case IDX_PLAY:
            if (!glCurrent) PlayTrack(0);
            else PlayAudio();
            break;

         case IDX_PAUSE:
            Pause();
            break;

         case IDX_STOP:
            StopPlaying();
            break;

         case IDX_NEXT:
            if (glTrack < glPlayIndex-1) PlayTrack(glTrack+1);
            else PlayTrack(0); /* Go back to the first track */
            break;

         case IDX_OPEN:
            glAddFiles = FALSE;
            Action(AC_Show, glFileDialog, NULL);
            break;

         case IDX_PLAYLIST:
            PlaylistWindow();
            break;
      }
   }
   else if ((Args->ActionID IS AC_Activate) AND (Args->ObjectID IS glFileDialog->UniqueID)) {
      STRING file, *files;
      WORD filecount;

      DPrintF("~Player", "Dialog box activated.");

      if ((glAddFiles IS TRUE) AND (glPlayIndex > 0)) {
         /*** Add all selected files to the playlist ***/

         filecount = 0;
         if ((GetField(glFileDialog, FID_SelectedFiles, FT_POINTER, &files) IS ERR_Okay) AND (files)) {
            for (filecount=0; files[filecount]; filecount++) {
               AddPlayFile(files[filecount]);
            }
         }

         if (filecount < 1) {
            if ((GetField(glFileDialog, FID_Location, FT_STRING, &file) IS ERR_Okay) AND (file)) {
               AddPlayFile(file);
            }
         }
      }
      else {

         /*** Add all selected files to the playlist ***/

         StopPlaying();
         glTrack = 0;

         if (glCurrent) {
            Action(AC_Free, glCurrent, NULL);
            glCurrent = NULL;
         }

         filecount = 0;
         ClearPlaylist();
         if ((GetField(glFileDialog, FID_SelectedFiles, FT_POINTER, &files) IS ERR_Okay) AND (files)) {
            for (filecount=0; files[filecount]; filecount++) {
               AddPlayFile(files[filecount]);
            }
         }

         /*** If one or fewer files were selected, automatically open and play the file ***/

         if (filecount <= 1) {
            if ((GetField(glFileDialog, FID_Location, FT_STRING, &file) IS ERR_Okay) AND (file)) {
               if (OpenFile(file) IS ERR_Okay) {
                  PlayAudio();
               }
            }
         }
         else PlayTrack(0);
      }

      StepBack();
   }
   else if ((Args->ActionID IS AC_Activate) AND (Args->ObjectID IS glPlaylistViewID)) {
      struct XMLTag **taglist;
      LONG tagindex;
      WORD track;

      /* The user has interacted with the playlist by double-clicking on an
      ** item.  The most recently activated item will be set in the ActiveTag field.
      */

      DPrintF("Player:","User has opened a playlist file.");

      if (GetFields(glPlaylistView, FID_ActiveTag|TLONG, &tagindex,
                                    FID_Tags|TPTR,       &taglist,
                                    TAGEND) IS ERR_Okay) {
         if (tagindex != -1) {
            if ((track = GetXMLAttributeInt(taglist[tagindex], "track"))) {
               PlayTrack(track-1);
            }
            else DPrintF("@Player:","Unable to identify the audio track for the selected item.");
         }
      }
   }
   else if ((Args->ActionID IS AC_Activate) AND (Args->ObjectID IS glPLAdd)) {
      glAddFiles = TRUE;
      acShow(glFileDialog);
   }
   else if ((Args->ActionID IS AC_Activate) AND (Args->ObjectID IS glPLDelete)) {
      struct XMLTag **taglist;
      LONG *tags;
      WORD i, track;

      /*** Delete all selected tags from our play list ***/

      DPrintF("Player:","Deleting selected tags from play list.");

      PrintPlaylist();

      if (GetFields(glPlaylistView, FID_SelectedTags|TPTR, &tags,
                                    FID_Tags|TPTR,         &taglist,
                                    TAGEND) IS ERR_Okay) {
         for (i=0; tags[i] != -1; i++) {
            if ((track = GetXMLAttributeInt(taglist[tags[i]], "track"))) {
               track--;
               RemovePlayFile(track);
               if (glTrack > track) glTrack--;
               else if (glTrack IS track) {
                  /*** The current track has just been deleted ***/
                  StopPlaying();
                  glTrack = 0;
                  if (glCurrent) {
                     acFree(glCurrent);
                     glCurrent = NULL;
                  }
               }
            }
         }

         /*** Clear the view, then recreate the view items ***/

         DPrintF("Player:","Clearing the view and recreating items.");
         acClear(glPlaylistView);

         for (i=0; i < glPlayIndex; i++) {
            AddViewXML(i);
         }
      }
   }

   return(ERR_Okay);
}

/*****************************************************************************
** Program: Timer()
*/

static ERROR PROGRAM_Timer(OBJECTPTR Self, struct acTimer *Args)
{
   LONG active;

   if (glStringWidth > glPanelWidth - (PANEL_MARGIN*2)) {
      if (!glScrollDirection) {
         /*** Scroll words to the left (reveal the end of the string) ***/
         glScrollX--;
         if ((PANEL_MARGIN*2) + glScrollX + glStringWidth < glPanelWidth) glScrollDirection = 1;
      }
      else {
         /*** Scroll words back to the right ***/
         glScrollX++;
         if (glScrollX IS 0) glScrollDirection = 0;
      }

      ActionMsg(AC_Draw, glOutputID, NULL);
   }

   if (glCurrent) {
      if (glCurrent->ClassID IS ID_MUSIC) {

      }
      else if (glCurrent->ClassID IS ID_SOUND) {
         if (GetField(glCurrent, FID_Active, FT_LONG, &active) != ERR_Okay) active = FALSE;

         if (glState IS STATE_PLAYING) {
            if (active IS FALSE) {
               DPrintF("Timer:","The audio sample is no longer active.");

               if (glTrack < glPlayIndex-1) PlayTrack(glTrack + 1);
               else {
                  /* If there are multiple files in the playlist, return
                  ** to the start by pre-loading the first track.  Otherwise,
                  ** set the position of the current audio sample back to 0.
                  */

                  if (glPlayIndex > 1) LoadTrack(0);
                  else StopPlaying();
               }
            }
         }
         else {
            if (active) DPrintF("@Timer:","The audio sample has been activated without my knowledge.");
         }
      }
   }

   return(ERR_Okay);
}

/*****************************************************************************
** Internal: AboutWindow()
*/

static void AboutWindow(void)
{
   static OBJECTID glAboutID = NULL;
   OBJECTPTR object, window;
   OBJECTID render_id;
   LONG textheight, left, top, bottom, right;

   if ((glAboutID) AND (CheckObjectExists(glAboutID, NULL) IS ERR_Okay)) {
      ActionMsg(AC_MoveToFront, glAboutID, NULL);
      ActionMsg(AC_Focus, glAboutID, NULL);
      return;
   }

   if (CreateObject(ID_WINDOW, NULL, &window, &glAboutID,
      FID_Title|TSTRING,      "About",
      FID_InsideWidth|TLONG,  300,
      FID_InsideHeight|TLONG, 150,
      FID_Icon|TSTRING,       "icons:items/info",
      FID_InsideBorder|TLONG, TRUE,
      FID_Quit|TLONG,         FALSE,
      TAGEND) IS ERR_Okay) {

      GetFields(window, FID_LeftMargin|TLONG,   &left,
                        FID_TopMargin|TLONG,    &top,
                        FID_BottomMargin|TLONG, &bottom,
                        FID_RightMargin|TLONG,  &right,
                        TAGEND);

      if (CreateObject(ID_RENDER, NULL, NULL, &render_id,
         FID_Owner|TLONG,    glAboutID,
         FID_XCoord|TLONG,   left,
         FID_YCoord|TLONG,   top,
         FID_XOffset|TLONG,  right,
         FID_YOffset|TLONG,  bottom,
         FID_Colour|TSTRING, "#ffffe0",
         TAGEND) IS ERR_Okay) {

         if (CreateObject(ID_TEXT, NULL, &object, NULL,
            FID_Owner|TLONG,    render_id,
            FID_YCoord|TLONG,   9,
            FID_Align|TLONG,    ALIGN_HORIZONTAL,
            FID_Face|TSTRING,   "helvete:12",
            FID_Colour|TSTRING, "#000000",
            FID_String|TSTRING, "Mini Player",
            TAGEND) IS ERR_Okay) {

            GetField(object, FID_TextHeight, FT_LONG, &textheight);

            CreateObject(ID_TEXT, NULL, NULL, NULL,
               FID_Owner|TLONG,    render_id,
               FID_XCoord|TLONG,   6,
               FID_XOffset|TLONG,  6,
               FID_YCoord|TLONG,   9 + textheight + 6,
               FID_Align|TLONG,    ALIGN_HORIZONTAL,
               FID_Face|TTRANS,    "[gl_fonts.face]",
               FID_Flags|TSTRING,  "!wordwrap",
               FID_Colour|TSTRING, "#000000",
               FID_String|TSTRING, "This program plays audio and digital music (module) files.  Support is provided for WAV, MP3, MOD, S3M, IT, XM and other formats if you have extra music extensions installed on your system.\n\nMini Player is the copyright of Rocklyte Systems © 2003.",
               TAGEND);
         }

         ActionMsg(AC_Show, render_id, NULL);
      }

      Action(AC_Show, window, NULL);

      ReleaseObject(window);
   }
}

/*****************************************************************************
** Internal: DrawOutput()
*/

static void DrawOutput(APTR Void, struct Render *Drawable, struct Bitmap *Bitmap)
{
   LONG clipleft, clipright;

   clipleft = Bitmap->ClipLeft;
   clipright = Bitmap->ClipRight;

   /*** Draw panel background and border ***/

   ActionTags(MT_DrawRGBRectangle, Bitmap, 0, 0, Drawable->Width, Drawable->Height, 255, 255, 255, TRUE);
   ActionTags(MT_DrawRGBRectangle, Bitmap, 0, 0, Drawable->Width, Drawable->Height, 128, 128, 128, FALSE);

   if (Bitmap->ClipLeft < 1) Bitmap->ClipLeft = 1;
   if (Bitmap->ClipRight > Drawable->Width - 1) Bitmap->ClipRight = Drawable->Width - 1;

   /*** Draw panel text ***/

   SetField(glFont, FID_Bitmap, FT_POINTER, Bitmap);
   SetField(glFont, FID_String, FT_STRING, glString);
   glFont->XCoord = PANEL_MARGIN + glScrollX;
   glFont->YCoord = 0;
   glFont->AlignFlags |= ALIGN_VERTICAL;
   glFont->AlignHeight = Drawable->Height;
   Action(AC_Draw, glFont, NULL);

   Bitmap->ClipLeft = clipleft;
   Bitmap->ClipRight = clipright;
}

/*****************************************************************************
** Internal: SetString()
** Short:    Sets the string that appears in the player's panel.
*/

static void SetString(STRING String)
{
   struct mtCalcStringWidth calc;

   StrCopy(String, glString, sizeof(glString));

   glScrollX = 0;
   glScrollDirection = 0;

   calc.String = glString;
   calc.Length = -1;
   if (Action(MT_CalcStringWidth, glFont, &calc) IS ERR_Okay) {
      glStringWidth = calc.Width;
   }
   else glStringWidth = 0;
}

/****************************************************************************/

#include "control.c"
#include "playlist.c"

